#include "subdivision.h"

Subdivision::Subdivision(QOpenGLFunctions_4_3_Core *f)
{
    this->f =f;
}

Subdivision::~Subdivision()
{
    delete edgeShader;
}

void Subdivision::init() {
    edgeShader = initEdgeComputeShaderProgram();
}

QVector<unsigned int> Subdivision::fillVector(unsigned int a, unsigned int b, unsigned int c,unsigned int d){
    QVector<unsigned int> x;
    x.push_back(a);
    x.push_back(b);
    x.push_back(c);
    x.push_back(d);
    return x;
}

QOpenGLShaderProgram *Subdivision::initEdgeComputeShaderProgram(){
    QString source = QLatin1String(":/subdivision-edge.compute");

    qDebug()<<"Compiling compute shader ...";
    QOpenGLShader *computeShader = new QOpenGLShader(QOpenGLShader::Compute);
    if(!computeShader->compileSourceFile(source)){
        qCritical()<<"Compute Shader"<<source<<"failed"<<computeShader->log();
        exit(5);
    }

    qDebug()<<"Adding compute shader ...";
    QOpenGLShaderProgram *shader = new QOpenGLShaderProgram();
    shader->addShader(computeShader);
    qDebug()<<"Linking compute shader ...";
    if(!shader->link()){
        qCritical()<<"Linking compute shader failed:"<<shader->log();
        exit(5);
    }
    qDebug()<<"Linking compute shader done";

    return shader;
}

Mesh *Subdivision::subdivide(Mesh *mesh) {
    QVector<QVector<unsigned int> > edgeIndices;
    QVector<unsigned int> vertexIndices;
    QVector<unsigned int> vertexIndicesOffsets;
    precomputeTables(mesh, edgeIndices, vertexIndices, vertexIndicesOffsets);
    runShader(mesh, edgeIndices);
    return NULL;
}

void Subdivision::precomputeTables(Mesh *mesh, QVector<QVector<unsigned int> > &edgeIndices_base,
                                   QVector<unsigned int> &vertexIndices, QVector<unsigned int> &vertexIndicesOffsets) {
    Mesh::Node root = mesh->getRootNode();

    int firstMeshEntryIndex = -1;
    bool meshExists = root.getFirstMeshIndex(firstMeshEntryIndex);
    if (!meshExists){
        qDebug()<<"No Mesh found. Abort subdiv table precomputation.";
        return;
    }
    qDebug()<<"Mesh found. Index:"<<firstMeshEntryIndex;

    //die drei zeilen machen das bild schwarz. wenn man davor returnt tuts.
    Mesh::MeshEntry *firstMeshEntry = mesh->getMeshEntry(firstMeshEntryIndex);
    QVector<unsigned int> ib = firstMeshEntry->getIndexBuffer();
    qDebug()<<"Index Buffer: "<<ib;

    QVector<Vertex> vb = firstMeshEntry->getVertexBuffer();
    //qDebug()<<"Vertex Buffer: "<<vb;

//compute edge table
    //Format: first two entries: edge vertices. last two entries: distant vertices.

    for (int i = 0; i < ib.length(); i+=3){
        //schaue alle dreiecke an

        Vertex x, y, z;
        unsigned int x_i = ib[i];
        unsigned int y_i = ib[i+1];
        unsigned int z_i = ib[i+2];

        //get vertices x,y,z from vertex buffer
        x = vb[x_i];//todo maybe check array length
        y = vb[y_i];
        z = vb[z_i];


        /*
         * Find edge xy in all other triangles, add to edgeIndices_base
         * Find edge yz
         * Find edge zx
         * - We assume that we have a closed surface, i.e. each triangle has exactly one adjacent triangle at each edge
         * - Each edge point will be computed twice. This allows us to implicitly know the adjacent points.
         */

        //find indices for edge point on xy-edge
        for (int j = 0; j < ib.length(); j+= 3){
            if (j != i){
                //search all following triangles for common edge
                Vertex a,b,c;
                unsigned int a_i, b_i, c_i;
                a_i = ib[j];
                b_i = ib[j+1];
                c_i = ib[j+2];

                a = vb[a_i];
                b = vb[b_i];
                c = vb[c_i];

                //get vertices for a,b,c from vertex buffer
                //comparisons: xy = ba, cb, ac.
                //yz = ba, cb, ac
                //zx = ba, cb, ac
                //TODO if oder else if?
                if (x.samePos(a) && y.samePos(c)){
                    edgeIndices_base.push_back(fillVector(x_i,y_i,z_i,b_i));
                }
                if (x.samePos(b) && y.samePos(a)){
                    edgeIndices_base.push_back(fillVector(x_i,y_i,z_i,c_i));
                }
                if (x.samePos(c) && y.samePos(b)){
                    edgeIndices_base.push_back(fillVector(x_i,y_i,z_i,a_i));
                }
            }
        }

        //find indices for edge point on yz-edge
        for (int j = 0; j < ib.length(); j+= 3){
            if (j != i){
                //search all following triangles for common edge
                Vertex a,b,c;
                unsigned int a_i, b_i, c_i;
                a_i = ib[j];
                b_i = ib[j+1];
                c_i = ib[j+2];

                a = vb[a_i];
                b = vb[b_i];
                c = vb[c_i];

                //yz = ba, cb, ac
                if (y.samePos(a) && z.samePos(c)){
                    edgeIndices_base.push_back(fillVector(y_i,z_i,x_i,b_i));
                }
                if (y.samePos(b) && z.samePos(a)){
                    edgeIndices_base.push_back(fillVector(y_i,z_i,x_i,c_i));
                }
                if (y.samePos(c) && z.samePos(b)){
                    edgeIndices_base.push_back(fillVector(y_i,z_i,x_i,a_i));
                }
            }
        }


        //find indices for edge point on xy-edge
        for (int j = 0; j < ib.length(); j+= 3){
            if (j != i){
                //search all following triangles for common edge
                Vertex a,b,c;
                unsigned int a_i, b_i, c_i;
                a_i = ib[j];
                b_i = ib[j+1];
                c_i = ib[j+2];

                a = vb[a_i];
                b = vb[b_i];
                c = vb[c_i];


                if (x.samePos(a) && z.samePos(b)){
                    edgeIndices_base.push_back(fillVector(x_i,z_i,y_i,c_i));
                }
                if (x.samePos(b) && z.samePos(c)){
                    edgeIndices_base.push_back(fillVector(x_i,z_i,y_i,a_i));
                }
                if (x.samePos(c) && z.samePos(a)){
                    edgeIndices_base.push_back(fillVector(x_i,z_i,y_i,b_i));
                }
            }
        }//quadratische Laufzeit ftw
       //Wichtig: Wir gehen davon aus, dass wir geschlossene Oberflächen haben, dh für jede Kante von einem Dreieck wird eine passende Kante bei einem anderen Dreieck gefunden.
    }

    qDebug()<<"Done with edge table. "<<edgeIndices_base.length()<<" edges found. Table: "<<edgeIndices_base;
    if (true){
        qDebug()<<"Table (long version):";
        for (int i = 0; i < edgeIndices_base.length(); i++){
            qDebug()<<"blub:"<<edgeIndices_base[i];
        }
    }

//compute vertex table
    //Format: First entry: original vertex. Other entries: adjacent vertices.
    QVector<QVector<unsigned int> > duplicates;//for debugging
    unsigned int offset = 0;
    for (unsigned int i = 0; i < vb.length(); i++){
        //hat evtl viel redundanz

        QVector<Vertex> adj_v;//helfer
        vertexIndicesOffsets.push_back(offset);
        vertexIndices.push_back(i);
        offset++;

        Vertex originalVertex = vb[i];

        QVector<unsigned int> d;

        for (int j = 0; j < ib.length(); j++){

            //suche verweise auf vertex im indexbuffer
            if (vb[ib[j]].samePos(originalVertex)){
                d.push_back(j);
                unsigned int i1, i2; //indices for neighbour vertices
                if(j%3==0){
                    i1 = ib[j+1];
                    i2 = ib[j+2];
                } else if(j%3==1){
                    i1 = ib[j-1];
                    i2 = ib[j+1];
                } else{
                    i1 = ib[j-2];
                    i2 = ib[j-1];
                }

                Vertex v1,v2; //neighbour vertices;
                v1 = vb[i1];
                v2 = vb[i2];
                //check if vertices where already used to influence this vertex
                //Note: Can't use contain. Contain uses equal (would have to implement equal, but we're only interested in the position here.
                bool found1 = false;
                bool found2 = false;
                for (int k = 0; k < adj_v.length(); k++){
                    if (adj_v[k].samePos(v1)) found1 = true;
                    if(adj_v[k].samePos(v2)) found2 = true;
                }
                if (!found1){
                    adj_v.push_back(v1);
                    vertexIndices.push_back(i1);
                    offset++;
                }
                if (!found2){
                    adj_v.push_back(v2);
                    vertexIndices.push_back(i2);
                    offset++;
                }
                //if one adjacent vertex is referenced by multiple different indices, only its first index will appear in the adjacent_indices list.1
            }
        }
        if (!duplicates.contains(d)) duplicates.push_back(d);
    }

    qDebug()<<"Done with vertex index table. ";

    qDebug()<<"Duplicates: ";
    for (int i = 0; i < duplicates.length(); i++){
        qDebug()<<duplicates[i];
    }

}

void Subdivision::runShader(Mesh *mesh, QVector<QVector<unsigned int> > &edgeIndices) {
    qDebug()<<"Running compute shader";

    Mesh::Node root = mesh->getRootNode();

    int firstMeshEntryIndex = -1;
    bool meshExists = root.getFirstMeshIndex(firstMeshEntryIndex);
    if (!meshExists){
        qDebug()<<"No Mesh found. Abort shader execution.";
        return;
    }
    qDebug()<<"Mesh found. Index:"<<firstMeshEntryIndex;

    Mesh::MeshEntry *firstMeshEntry = mesh->getMeshEntry(firstMeshEntryIndex);
    GLuint vb_handle = firstMeshEntry->VB_handle;
    QVector<Vertex> vb = firstMeshEntry->getVertexBuffer();

    edgeShader->bind();
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, vb_handle);

    GLuint *edgeIDs = new GLuint[edgeIndices.size() * 4];
    for (uint i = 0; i < edgeIndices.size(); i++) {
        for (uint j = 0; j < 4; j++) {
            edgeIDs[4 * i + j] = edgeIndices[i][j];
        }
    }

    // Create an input buffer and set it to the value of offset
    GLuint indicesID;
    f->glGenBuffers(1, &indicesID);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, indicesID);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, edgeIndices.size() * sizeof(GLuint) * 4, &edgeIDs[0], GL_DYNAMIC_DRAW);
    delete edgeIDs;

    // Create an output buffer
    GLuint bufferID;
    f->glGenBuffers(1, &bufferID);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, bufferID);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, edgeIndices.size() * sizeof(Vertex), NULL, GL_DYNAMIC_DRAW);

    // Run the shader
    f->glDispatchCompute(edgeIndices.size(), 1, 1);

    // Wait for the shader to complete and the data to be written back to the global memory
    f->glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

    // Unbind the buffers from the shader
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, 0);

    // Map the output buffer so we can read the results
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, bufferID);
    Vertex *ptr;
    ptr = (Vertex *) f->glMapBuffer(GL_SHADER_STORAGE_BUFFER, GL_WRITE_ONLY);
    for (int i = 0; i < edgeIndices.size(); i++) {
        qDebug() << ptr[i].pos;
    }
    f->glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);

    // Delete the buffers the free the resources
    f->glDeleteBuffers(1, &indicesID);
    f->glDeleteBuffers(1, &bufferID);

    edgeShader->release();
}
